#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <time.h>
#include <pthread.h>
#include <signal.h>
#include "sched.h"
#include "list.h"
#include "task.h"


#define N_PRIORITY_LEVELS   5

linked_list ready_tasks[N_PRIORITY_LEVELS];

task_tcb *current_task;

typedef struct _task_release_event {
    task_tcb *task;
    unsigned long release_time;
    unsigned int period;
} task_release_event;


linked_list task_release_timer;

unsigned long int ms_time;

pthread_mutex_t time_mutex;
pthread_mutex_t ready_list_mutex;

unsigned int sched_get_time_in_ms(void)
{
    unsigned long int return_me;
    pthread_mutex_lock(&time_mutex);
    return_me = ms_time;
    pthread_mutex_unlock(&time_mutex);
    return return_me;
}

void insert_timer_event(task_release_event *new_event, unsigned int new_release_time)
{
    task_release_event *current_event;

    new_event->release_time = new_release_time;
    task_release_timer.current_element = task_release_timer.first_element;

    current_event = (task_release_event*) list_get_current(&task_release_timer);
    while(current_event != NULL){
        if (current_event->release_time > new_release_time)
        {
            list_add_element_before_current(&task_release_timer, (void*)new_event);
            return;
        }
        list_get_next(&task_release_timer);
        current_event = (task_release_event*) list_get_current(&task_release_timer);
    }
    list_add_element(&task_release_timer, (void*)new_event);
}

void add_periodic_task_release_event(task_tcb *task, unsigned int period)
{
    task_release_event *new_event;
    unsigned long new_release_time = sched_get_time_in_ms() + period;


    new_event = malloc(sizeof(task_release_event));

    new_event->task = task;
    new_event->period = period;

    insert_timer_event(new_event, new_release_time);
}

int sched_register_task(task_tcb *task, unsigned int period)
{
    if(task == NULL)
    {
        return -1;
    }

    if(period > 0)
    {
        add_periodic_task_release_event(task, period);
    }
    else {
        pthread_mutex_lock(&ready_list_mutex);
        list_add_element(&(ready_tasks[task->priority]), (void*) task);
        pthread_mutex_unlock(&ready_list_mutex);
    }
    return 0;
}

void sched_swap_tasks(void)
{
    int priority;
    pthread_mutex_lock(&ready_list_mutex);
    for (priority = N_PRIORITY_LEVELS - 1; priority >= 0; --priority)
    {
        if (ready_tasks[priority].n_elements == 0)
        {
            continue;
        }
        current_task = (task_tcb*)list_get_next(&ready_tasks[priority]);
        while (task_get_finished(current_task) == 1)
        {
            list_remove_by_content(&ready_tasks[priority], (void*) current_task);
            task_set_finished(current_task, 0);
            current_task = (task_tcb*)list_get_next(&ready_tasks[priority]);
        }
        if (current_task != NULL)
        {
            break;
        }
    }
    pthread_mutex_unlock(&ready_list_mutex);
}

void ensure_no_deadline_miss(void)
{
    // This function manipulates the list directly in order to not
    // ruin the schedulers order of tasks 
    int schedulable = 0;
    int lowest_deadline, lowest_crit_level, priority;
    unsigned int wcet_sum = 0;
    list_element *current_loop_element;
    task_tcb *current_loop_tcb, *lowest_crit_task;

    while (!schedulable)
    {
        schedulable = 1;
        for (priority = N_PRIORITY_LEVELS - 1; priority >= 0; --priority)
        {
            if (ready_tasks[priority].n_elements == 0)
            {
                continue;
            }
            current_loop_element = ready_tasks[N_PRIORITY_LEVELS-1].first_element;
            current_loop_tcb = (task_tcb*)(current_loop_element->content);
            lowest_deadline = current_loop_tcb->deadline;
            lowest_crit_level = current_loop_tcb->criticality_level;
            while (current_loop_tcb != NULL)
            {
                wcet_sum += task_get_wcet(current_loop_tcb);
                if (current_loop_tcb->deadline < lowest_deadline)
                {
                    lowest_deadline = current_loop_tcb->deadline;
                }
                if (current_loop_tcb->criticality_level < lowest_crit_level)
                {
                    lowest_crit_task = current_loop_tcb;
                    lowest_crit_level = current_loop_tcb->criticality_level;
                }
                current_loop_element = current_loop_element->next; 
                current_loop_tcb = (task_tcb*)(current_loop_element->content);
            }
            if (lowest_deadline < wcet_sum)
            {
                schedulable = 0;
            }
        }
        if (!schedulable)
        {
            if (task_use_next_variant(lowest_crit_task) == 1)
            {
                // No more task variants. Abort task
            }
        }
    }
}

void *timer_thread(void *args)
{
    task_release_event *next_timer_event;
    unsigned long int current_time;
    task_release_timer.current_element = task_release_timer.first_element;
    next_timer_event = (task_release_event*)list_remove_current(&task_release_timer);
    if (next_timer_event != NULL)
    {
        while(1) 
        {
            current_time = sched_get_time_in_ms();
            pthread_mutex_lock(&ready_list_mutex);
            if (next_timer_event->task->estimator_preproc != NULL)
            {
                next_timer_event->task->estimator_preproc(next_timer_event->task->preproc_param, next_timer_event->task->shared->preproc_results);
                //next_timer_event->task->estimator_preproc(next_timer_event->task->preproc_param, &(next_timer_event->task->shared->preproc_results));
            }
            list_add_element(&(ready_tasks[next_timer_event->task->priority]), (void*) next_timer_event->task);
            /*ensure_no_deadline_miss();*/
            pthread_mutex_unlock(&ready_list_mutex);
            insert_timer_event(next_timer_event, current_time + next_timer_event->period);
            task_release_timer.current_element = task_release_timer.first_element;
            next_timer_event = (task_release_event*)list_remove_current(&task_release_timer);
            usleep(1000 * (next_timer_event->release_time - current_time));
        }
    }
    pthread_exit(NULL);
}

void *clock_thread(void *args)
{
    while(1) {
        pthread_mutex_lock(&time_mutex);
        ms_time++;
        pthread_mutex_unlock(&time_mutex);
        usleep(1000);
    }
}

void sched_init(void)
{
    int i;

    list_init(&task_release_timer);
    for (i = 0; i < N_PRIORITY_LEVELS; ++i)
    {
        list_init(&(ready_tasks[i]));
    }

    ms_time = 0;
    pthread_mutex_init(&time_mutex, NULL);
}

void sched_start(void)
{
    pthread_t timer, clock;

    pthread_mutex_init(&ready_list_mutex, NULL);

    if (pthread_create(&clock, NULL, clock_thread, NULL))
    {
        perror("pthread_create could not start timer thread");
    }
    
    if (pthread_create(&timer, NULL, timer_thread, NULL))
    {
        perror("pthread_create could not start timer thread");
    }

    // Ignore SIGCHLD so child processes don't become deprecated 
    signal(SIGCHLD, SIG_IGN);
    
    while(1) {
        
        task_suspend(current_task);
        sched_swap_tasks();
        task_start(current_task);

        usleep(10000);
    }
}

