#include <unistd.h>
#include <stdlib.h>
#include <pthread.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/signal.h>
#include <stdio.h>
#include <fcntl.h>
#include <errno.h>

#include "task.h"
#include "list.h"

int task_init(task_tcb *new_task, char *name, unsigned int priority, unsigned int deadline, unsigned int criticality_level, void (*estimator_preproc)(void* args, void* result), void *preproc_param)
{
    int mutex_fd, rc;
    pthread_mutexattr_t mutex_attr;
    char shared_mem_name[TASK_MAX_NAME_SIZE + 2] = ""; // Initialized to empty string to make valgrind happy
    if (new_task == NULL)
    {
        return -1;
    }
    if (name == NULL)
    {
        return -1;
    }

    new_task->task_variants = malloc(sizeof(linked_list));
    list_init(new_task->task_variants);
    new_task->n_task_variants = 0;
    new_task->priority = priority;
    new_task->deadline = deadline;
    new_task->criticality_level = criticality_level;
    new_task->estimator_preproc = estimator_preproc;
    new_task->preproc_param = preproc_param;

    strncpy(new_task->name, name, TASK_MAX_NAME_SIZE);
    if (strlen(name) >= TASK_MAX_NAME_SIZE) // Not strictly necessary, but looks good
    {
        new_task->name[TASK_MAX_NAME_SIZE] = '\0';
    }
    shared_mem_name[0] = '/';
    strcat(shared_mem_name, new_task->name);
    
    mutex_fd = shm_open(new_task->name, O_CREAT | O_RDWR | O_TRUNC, S_IRWXU | S_IRWXG);
    if (mutex_fd < 0)
    {
        perror("shm_open failed");
        return -1;
    }

    //rc = ftruncate(mutex_fd, sizeof(pthread_mutex_t));
    rc = ftruncate(mutex_fd, sizeof(task_shared_data));
    if (rc != 0)
    {
        perror("ftruncate failed");
        return -1;
    }

    //new_task->mutex = (pthread_mutex_t*) mmap(NULL, sizeof(pthread_mutex_t),
    //        PROT_READ | PROT_WRITE, MAP_SHARED, mutex_fd, 0);
    new_task->shared = (task_shared_data*) mmap(NULL, sizeof(task_shared_data),
            PROT_READ | PROT_WRITE, MAP_SHARED, mutex_fd, 0);
    if (new_task == MAP_FAILED)
    {
        perror("mmap failed");
        return -1;
    }

    new_task->shared->pid = -1;
    new_task->shared->finished = 0;
    //new_task->shared->preproc_results = 0;
    
    rc = pthread_mutexattr_init(&mutex_attr);
    if (rc != 0)
    {
        perror("mutexattr_init failed");
        return -1;
    }

    rc = pthread_mutexattr_setpshared(&mutex_attr, PTHREAD_PROCESS_SHARED);
    if (rc != 0)
    {
        perror("mutexattr_setpshared failed");
        return -1;
    }

    //new_task->mutex = (pthread_mutex_t*) malloc(sizeof(pthread_mutex_t));
    rc = pthread_mutex_init(&(new_task->shared->mutex), &mutex_attr);
    if (rc != 0)
    {
        perror("mutex_init failed");
        return -1;
    }

    rc = pthread_mutexattr_destroy(&mutex_attr);
    if (rc != 0)
    {
        perror("mutexattr_destroy_failed");
        return -1;
    }

    return 0;
}

int task_add_variant(task_tcb *task, task_variant *task_var)
{
    if (task == NULL)
    {
        return -1;
    }
    if (task_var == NULL)
    {
        return -1;
    }
    list_add_element(task->task_variants, (void*) task_var);
    task->n_task_variants++;

    return 0;
}

int task_variant_init(task_variant *task_var, void (*task_routine)(void *), void *task_routine_param, 
                      unsigned int (*wcet_estimator)(void *), void *estimator_param)
{
    if (task_var == NULL)
    {
        return -1;
    }
    task_var->task_routine = task_routine;
    task_var->task_routine_param = task_routine_param;
    task_var->wcet_estimator = wcet_estimator;
    task_var->estimator_param = estimator_param;

    return 0;
}

unsigned int task_get_wcet(task_tcb *task)
{
    if (task == NULL)
    {
        // TODO: Maybe not ideal to return an error code of -1 as an unsigned int...
        return -1;
    }
    
    task_variant *current_var = (task_variant*) list_get_current(task->task_variants);
    //if (task->shared->preproc_results == 0)
    if (task->shared->preproc_results == NULL)
    {
        return current_var->wcet_estimator(current_var->estimator_param);
    }
    return current_var->wcet_estimator(task->shared->preproc_results);
    //return current_var->wcet_estimator(&(task->shared->preproc_results));
}

int task_use_next_variant(task_tcb *task)
{
    if (task == NULL)
    {
        return -1;
    }
    pthread_mutex_lock(&(task->shared->mutex));
    if (task->shared->pid != -1)
    {
        /* task is running */
        kill(task->shared->pid, SIGKILL);
        task->shared->pid = -1;
    }
    pthread_mutex_unlock(&(task->shared->mutex));

    if (list_get_next(task->task_variants) == NULL)
    {
        return 1;
    }
    return 0;
}

int task_start(task_tcb *task)
{
    pid_t child_pid;
    task_variant *current_var;
    
    if (task == NULL)
    {
        return -1;
    }

    pthread_mutex_lock(&(task->shared->mutex));


    if (task->shared->pid != -1)
    {
        //Task has a process, resume it
        kill(task->shared->pid, SIGCONT);
        pthread_mutex_unlock(&(task->shared->mutex));
    }
    else {
        //Task has no process, start a new
        child_pid = fork();
        if (child_pid < 0)
        {
            perror("fork failed");
            return -1;
        }
        else {
            if (child_pid == 0) //Child process
            {
                pthread_mutex_lock(&(task->shared->mutex)); //To ensure that the parent sets the pid before task runs
                pthread_mutex_unlock(&(task->shared->mutex));
                current_var = (task_variant*) list_get_current(task->task_variants);
                current_var->task_routine(current_var->task_routine_param);
                // Task is done
                pthread_mutex_lock(&(task->shared->mutex));
                task->shared->finished = 1;
                task->shared->pid = -1;
                pthread_mutex_unlock(&(task->shared->mutex));
                exit(0);
            }
            else {
                task->shared->pid = child_pid;
                pthread_mutex_unlock(&(task->shared->mutex));
            }
        }
    }
    return 0;
}

int task_get_finished(task_tcb *task)
{
    int return_me = 0;
    if (task == NULL)
    {
        return -1;
    }
    pthread_mutex_lock(&(task->shared->mutex));
    if (task->shared->finished)
    {
        return_me = 1;
    }
    pthread_mutex_unlock(&(task->shared->mutex));
    return return_me;
}

int task_set_finished(task_tcb *task, int new_value)
{
    if (task == NULL)
    {
        return -1;
    }
    if (new_value != 0 && new_value != 1)
    {
        return -1;
    }
    pthread_mutex_lock(&(task->shared->mutex));
    task->shared->finished = new_value;
    pthread_mutex_unlock(&(task->shared->mutex));
    return 0;
}

int task_suspend(task_tcb *task)
{
    if (task == NULL)
    {
        return -1;
    }

    pthread_mutex_lock(&(task->shared->mutex));
    if (task->shared->pid == -1)
    {
        // Task is not running
        pthread_mutex_unlock(&(task->shared->mutex));
        return -1;
    }
    kill(task->shared->pid, SIGSTOP);
    pthread_mutex_unlock(&(task->shared->mutex));

    return 0;
}

int task_resume(task_tcb *task)
{
    if (task == NULL)
    {
        return -1;
    }

    pthread_mutex_lock(&(task->shared->mutex));
    if (task->shared->pid == -1)
    {
        // Task is not running
        pthread_mutex_unlock(&(task->shared->mutex));
        return -1;
    }
    kill(task->shared->pid, SIGCONT);
    pthread_mutex_unlock(&(task->shared->mutex));
    
    return 0;
}

int task_destroy(task_tcb *task)
{
    if (task == NULL)
    {
        return -1;
    }

    if (task->shared->pid != -1)
    {
        kill(task->shared->pid, SIGKILL);
    }
    pthread_mutex_destroy(&(task->shared->mutex));
    shm_unlink(task->name);

    list_destroy(task->task_variants);
    free(task->task_variants);
    return 0;
}
